/**************************************************************************
*  Copyright (c) 2020-2022 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <stdio.h>
#include "tal.h"
#include "terminal.h"
#include "dhrystone.h"
#include "memtest.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define CAN_RX_RING_BUFFER_SIZE  128
#define CAN_TX_RING_BUFFER_SIZE  128

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/* 
 * Some TASK variables like stack and task control block.
 */
static OS_STACK (StartStack,  TASK_START_STK_SIZE);
static OS_STACK (LEDStack,    TASK_LED_STK_SIZE);
static OS_STACK (CANStack,    TASK_CAN_STK_SIZE);

static OS_STACK (AStack,      TASK_A_STK_SIZE);
static OS_STACK (BStack,      TASK_B_STK_SIZE);
static OS_STACK (CStack,      TASK_C_STK_SIZE);

static OS_TCB   TCBStartTask;
static OS_TCB   TCBLed;
static OS_TCB   TCBCan;

static OS_TCB   TCBA;
static OS_TCB   TCBB;
static OS_TCB   TCBC;

static uint8_t bABCTestRunning = 0;

static TAL_CAN_OBJECT   CAN2RxRingBuffer[CAN_RX_RING_BUFFER_SIZE];
static TAL_CAN_OBJECT   CAN2TxRingBuffer[CAN_TX_RING_BUFFER_SIZE];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

#if defined(MEMORY_INIT)
/*************************************************************************/
/*  MemTest                                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 = OK / -1 = Error                                          */
/*************************************************************************/
int MemTest (void)
{
   /*
    * The data bus width must be set in memtest.h,
    * datum must be set.
    */
   #define BASE_ADDRESS    (volatile datum *) 0x80000000
   #define NUM_BYTES       (32 * 1024 * 1024)
   
   
   if ( (memTestDataBus(BASE_ADDRESS) != 0)                  ||
        (memTestAddressBus(BASE_ADDRESS, NUM_BYTES) != NULL) ||
        (memTestDevice(BASE_ADDRESS, NUM_BYTES) != NULL)     )
   {
      return (-1);
   }
   else
   {
      return (0);
   }
   
} /* MemTest() */
#endif /* defined(MEMORY_INIT) */

/*************************************************************************/
/*  OutputBootMessage                                                    */
/*                                                                       */
/*  Output boot message.                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputBootMessage (void)
{
   const char ResetScreen[] = { 0x1B, 'c', 0 };
   
   term_printf("%s", ResetScreen);
   OS_TimeDly(50);   
   
   term_printf("\r\n");
   term_printf("*********************************\r\n");
   term_printf("  Project: %s\r\n", PROJECT_NAME);
   term_printf("  Board  : %s\r\n", TAL_BOARD);
   term_printf("  Version: v%s\r\n", PROJECT_VER_STRING);
   term_printf("  Build  : "__DATE__ " " __TIME__"\r\n");
   term_printf("*********************************\r\n");
   term_printf("\r\n");
   
} /* OutputBootMessage */

/*************************************************************************/
/*  OutputVersionInfo                                                    */
/*                                                                       */
/*  Output version information.                                          */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputVersionInfo (void)
{
   term_printf("*** Version ***\r\n");
   term_printf("Board: %s\r\n", TAL_BOARD);
   term_printf("CPU  : %s\r\n", TAL_CPU);
   term_printf("OS   : %s v%s\r\n", OS_NAME, OS_VER_STRING);
   term_printf("TAL  : v%s\r\n", TAL_CORE_VER_STRING);
   term_printf("\r\n");
   
} /* OutputVersionInfo */

/*************************************************************************/
/*  OutputFrequencyInfo                                                  */
/*                                                                       */
/*  Output frequency information.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputFrequencyInfo (void)
{
   term_printf("*** Frequency info ***\r\n");
   term_printf("CPU: %3d MHz\r\n", tal_CPUGetFrequencyCPU() / 1000000); 
   term_printf("AHB: %3d MHz\r\n", tal_CPUGetFrequencyAHB() / 1000000); 
   term_printf("IPG: %3d MHz\r\n", tal_CPUGetFrequencyIPG() / 1000000); 
   term_printf("USB: %3d MHz\r\n", tal_CPUGetFrequencyUSB() / 1000000); 
   term_printf("OSC: %3d MHz\r\n", tal_CPUGetFrequencyOSC() / 1000000); 
   term_printf("\r\n");
   
} /* OutputFrequencyInfo */

/*************************************************************************/
/*  OutputCPULoad                                                        */
/*                                                                       */
/*  Output load information.                                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OutputCPULoad (void)
{
   term_printf("CPU load: %d%%\r\n\r\n", OS_StatGetCPULoad());
   
} /* OutputCPULoad */

/*************************************************************************/
/*  OutputUsageInfo                                                      */
/*                                                                       */
/*  Output usage information.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputUsageInfo (void)
{
   term_printf("*** Usage ***\r\n");
   term_printf("h: output this info\r\n");
   term_printf("?: output this info\r\n");
   term_printf("c: output cpu load info\r\n");
   term_printf("f: output frequency info\r\n");
   term_printf("r: output runtime stack info\r\n");
   term_printf("t: output task info\r\n");
   term_printf("m: output memory info\r\n");
   term_printf("v: output version info\r\n");
   term_printf("x: reboot\r\n");
   term_printf("\r\n");
   term_printf("a: start round robin ABC task wait test\r\n");
   term_printf("A: stop round robin ABC task wait test\r\n");
   term_printf("\r\n");
   
} /* OutputUsageInfo */

/*************************************************************************/
/*  ATask                                                                */
/*                                                                       */
/*  This is the A task.                                                  */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void ATask (void *p)
{
   (void)p;
   
   while (1 == bABCTestRunning)
   {
      term_putchar('A');   
      OS_TimeDly(1000);
   }

} /* ATask */

/*************************************************************************/
/*  BTask                                                                */
/*                                                                       */
/*  This is the B task.                                                  */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void BTask (void *p)
{
   (void)p;
   
   while (1 == bABCTestRunning)
   {
      term_putchar('B');   
      OS_TimeDly(1000);
   }

} /* BTask */

/*************************************************************************/
/*  CTask                                                                */
/*                                                                       */
/*  This is the C task.                                                  */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void CTask (void *p)
{
   (void)p;
   
   while (1 == bABCTestRunning)
   {
      term_putchar('C');   
      OS_TimeDly(1000);
   }
   
} /* CTask */

/*************************************************************************/
/*  LEDTask                                                              */
/*                                                                       */
/*  This is the LED task.                                                */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void LEDTask (void *p)
{
   (void)p;
   
   while (1)
   {
      tal_LEDSet(LED16);
      OS_TimeDly(TASK_LED_DELAY_MS);
      tal_LEDClear(LED16);
      OS_TimeDly(TASK_LED_DELAY_MS);
   }

} /* LEDTask */

/*************************************************************************/
/*  CANTask                                                              */
/*                                                                       */
/*  This is the CAN1 task.                                               */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/

static TAL_CAN_DCB  CANPort;

static void CANTask (void *p)
{
   TAL_RESULT        Error;
   TAL_CAN_SETTINGS  CANSettings;   
   TAL_CAN_OBJECT    CANObject;

   (void)p;

   /* Init the Device Control Block */      
   tal_CANInitDCB(&CANPort, TAL_CAN_PORT_2);
   
   /* Set Rx and Tx ring buffer */
   tal_CANSetRingBuffer(&CANPort, TAL_CAN_BUFFER_RX, CAN2RxRingBuffer, sizeof(CAN2RxRingBuffer)); 
   tal_CANSetRingBuffer(&CANPort, TAL_CAN_BUFFER_TX, CAN2TxRingBuffer, sizeof(CAN2TxRingBuffer)); 

   /* Prepare communication settings */
   CANSettings.dBitRate = TAL_CAN_BR_500K; 

   /* Now, the port can be opened */   
   tal_CANOpen(&CANPort, &CANSettings);  
   
   tal_CANIdentRegister(&CANPort, TAL_CAN_ID_RX_ALL, TAL_CAN_TYPE_STD_ID);
   tal_CANIdentRegister(&CANPort, TAL_CAN_ID_RX_ALL, TAL_CAN_TYPE_EXT_ID);

   while (1)
   {
      Error = tal_CANMsgGetWait(&CANPort, &CANObject, TASK_CAN_DELAY_MS);
      if (TAL_OK == Error)
      {
         /* If received STD identifier send EXT, and vice versa */ 
         if (CANObject.bFlags & TAL_CAN_OBJECT_FLAGS_STD_ID)
         {
            CANObject.bFlags = TAL_CAN_OBJECT_FLAGS_EXT_ID;
         }
         else
         {
            CANObject.bFlags = TAL_CAN_OBJECT_FLAGS_STD_ID;
         }
         
         /* Send object */
         tal_CANMsgSend(&CANPort, &CANObject);
      }
      else
      {
         if( (TAL_ERR_CAN_OVERFLOW       == Error) || 
             (TAL_ERR_CAN_OVERFLOW_EMPTY == Error) )
         {
            tal_CANClearOverflow(&CANPort);
         }

         OS_TimeDly(TASK_CAN_DELAY_MS);
      } /* end if (TAL_OK == Error) */ 
   } /* end while (1) */
   
} /* CANTask */

/*************************************************************************/
/*  StartTask                                                            */
/*                                                                       */
/*  This is the Start task.                                              */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StartTask (void *p)
{
   (void)p;
   
   /*
    * The StartTask will be used to start all other tasks in the system.
    * At the end the priority will be set to "IDLE" priority.
    */
    
   OS_SysTickStart();   /* Start the System ticker */    
   OS_StatEnable();     /* Enable the statistic function  */

   term_Start();        /* Start the Terminal functionality */
   
   /*
    * Output startup messages
    */
   OutputBootMessage();
   OutputFrequencyInfo();
   OutputUsageInfo();

   /* Create the LED task */
   OS_TaskCreate(&TCBLed, LEDTask, NULL, TASK_LED_PRIORITY,
                 LEDStack, sizeof(LEDStack), 
                 "LEDTask");

   /* Create the CAN task */
   OS_TaskCreate(&TCBCan, CANTask, NULL, TASK_CAN_PRIORITY,
                 CANStack, sizeof(CANStack), 
                 "CANTask");

   /*
    * Dhrystone test
    */
#if 0
   OS_TimeDly(500);
   main_dhry();
#endif
   
   /*
    * Memory test
    */           
#if defined(MEMORY_INIT) && !defined(__SDRAM__)
   term_printf("Memory test... ");
   OS_TimeDly(100);
   if (0 == MemTest())
   {
      term_printf("OK\r\n");
   }
   else
   {
      term_printf("ERROR\r\n");
   }
#endif   
                   
   OS_TaskChangePriority(TASK_START_PRIORITY_IDLE);

   while (1)
   {
      OS_TimeDly(TASK_START_DELAY_MS);
   }

} /* StartTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
int main (void)
{
   /* 
    * Init the Tiny Abstraction Layer 
    */
   tal_Init();
   
   /* 
    * Create the StartTask.
    * The StartTask is the one and only task which 
    * will start all other tasks of the system.
    */
   OS_TaskCreate(&TCBStartTask, StartTask, NULL, TASK_START_PRIORITY,
                 StartStack, sizeof(StartStack), 
                 "StartTask");
   
   /* 
    * OSStart must be the last function here.
    *  
    * Fasten your seatbelt, engine will be started...
    */        
   OS_Start();
   
   /*
    * This return here make no sense.
    * But to prevent the compiler warning:
    *    "return type of 'main' is not 'int'
    * We use an int as return :-)
    */ 
   return(0); /*lint !e527*/
} /* main */

/*************************************************************************/
/*  term_RxCallback                                                      */
/*                                                                       */
/*  Will be called from TermTask in case a char is received.             */
/*                                                                       */
/*  In    : bData                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void term_RxCallback (uint8_t bData)
{
   switch (bData)
   {
      case 'h':
      case '?':
      {
         OutputUsageInfo();
         break;
      } 
           
      case 'c':
      {
         OutputCPULoad();
         break;
      }

      case 'f':
      {
         OutputFrequencyInfo();
         break;
      }

      case 'r':
      {
         OS_OutputRuntimeStackInfo();
         break;
      }
      
      case 't':
      {
         OS_OutputTaskInfo();
         break;
      }
      
      case 'm':
      {
         tal_MEMOutputMemoryInfo();
         break;
      }
      
      case 'v':
      {
         OutputVersionInfo();
         break;
      }
      
      case 'x':
      {
         tal_CPUReboot();
         break;
      }
      
      /****************************************************/
      
      case 'a':
      {
         if (0 == bABCTestRunning)
         {
            bABCTestRunning = 1;
            
            /* Create the A,B and C task */
            OS_TaskCreate(&TCBA, ATask, NULL, TASK_A_PRIORITY,
                          AStack, sizeof(AStack), "ATask");

            OS_TaskCreate(&TCBB, BTask, NULL, TASK_B_PRIORITY,
                          BStack, sizeof(BStack), "BTask");

            OS_TaskCreate(&TCBC, CTask, NULL, TASK_C_PRIORITY,
                          CStack, sizeof(CStack), "CTask");
         }
         break;
      }
      
      case 'A':
      {
         if (1 == bABCTestRunning)
         {
            bABCTestRunning = 0;
            OS_TimeDly(1000);
            term_printf("\r\n\r\n");
         }
         break;
      }  
      
      case '1':
      {      
         TAL_CAN_OBJECT CANObject;
         
         CANObject.dIdentifier = 0x123;
         CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_STD_ID;
         CANObject.bDLC        = 8;
         CANObject.Data[0]     = 0x11;
         CANObject.Data[1]     = 0x22;
         CANObject.Data[2]     = 0x33;
         CANObject.Data[3]     = 0x44;
         CANObject.Data[4]     = 0x55;
         CANObject.Data[5]     = 0x66;
         CANObject.Data[6]     = 0x77;
         CANObject.Data[7]     = 0x88;
         
         tal_CANMsgSend(&CANPort, &CANObject);
         break;
      }   
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (bData) */
   
} /* term_RxCallback */

/*************************************************************************/
/*  HardFault_Handler                                                    */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void HardFault_Handler (void)
{

#if defined(__DEBUG__)
   while (1)
   {
      __asm__ ("nop");
   }
#endif

   tal_CPUReboot();

} /* HardFault_Handler */

/*** EOF ***/
